@@ -11,7 +11,9 @@ from agoraapi.transaction.v3 import transaction_service_pb2 as tx_pb_v3, transac
 from agoraapi.transaction.v4 import transaction_service_pb2 as tx_pb_v4, transaction_service_pb2_grpc as tx_pb_grpc_v4
 from kin_base import transaction_envelope as te
 
+from agora import solana
 from agora.cache.cache import LRUCache
+from agora.client.utils import _generate_token_account
 from agora.error import BlockchainVersionError, AccountExistsError, AccountNotFoundError, TransactionRejectedError, \
     TransactionErrors, Error, InsufficientBalanceError, PayerRequiredError, NoSubsidizerError, AlreadySubmittedError, \
     BadNonceError
@@ -30,10 +32,10 @@ _SERVICE_CONFIG_CACHE_KEY = b'GetServiceConfig'
 class SubmitTransactionResult:
     def __init__(self, tx_id: Optional[bytes] = None,
                  invoice_errors: Optional[List[model_pb_v3.InvoiceError]] = None,
-                 tx_error: Optional[TransactionErrors] = None):
+                 errors: Optional[TransactionErrors] = None):
         self.tx_id = tx_id if tx_id else bytes(32)
         self.invoice_errors = invoice_errors if invoice_errors else []
-        self.tx_error = tx_error
+        self.errors = errors
 
 
 class InternalClient:
@@ -202,7 +204,7 @@ class InternalClient:
             elif resp.result == tx_pb_v3.SubmitTransactionResponse.Result.INVOICE_ERROR:
                 result.invoice_errors = resp.invoice_errors
             elif resp.result == tx_pb_v3.SubmitTransactionResponse.Result.FAILED:
-                result.tx_error = TransactionErrors.from_result(resp.result_xdr)
+                result.errors = TransactionErrors.from_result(resp.result_xdr)
             elif resp.result != tx_pb_v3.SubmitTransactionResponse.Result.OK:
                 raise Error(f'unexpected result from agora: {resp.result}')
 
@@ -220,6 +222,8 @@ class InternalClient:
             transaction payer.
         """
 
+        token_account_key = _generate_token_account(private_key)
+
         def _create():
             nonlocal subsidizer
 
@@ -244,19 +248,19 @@ class InternalClient:
                 [
                     system.create_account(
                         subsidizer_id,
-                        private_key.public_key,
+                        token_account_key.public_key,
                         token_program,
                         min_balance_resp.lamports,
                         token.ACCOUNT_SIZE,
                     ),
                     token.initialize_account(
-                        private_key.public_key,
+                        token_account_key.public_key,
                         PublicKey(service_config_resp.token.value),
                         private_key.public_key,
                         token_program,
                     ),
                     token.set_authority(
-                        private_key.public_key,
+                        token_account_key.public_key,
                         private_key.public_key,
                         token.AuthorityType.CloseAccount,
                         token_program,
@@ -265,7 +269,7 @@ class InternalClient:
                 ]
             )
             transaction.set_blockhash(recent_blockhash_resp.blockhash.value)
-            transaction.sign([private_key])
+            transaction.sign([private_key, token_account_key])
             if subsidizer:
                 transaction.sign([subsidizer])
 
@@ -336,18 +340,21 @@ class InternalClient:
         return TransactionData(tx_id, TransactionState.from_proto_v4(resp.state))
 
     def submit_solana_transaction(
-        self, tx_bytes: bytes, invoice_list: Optional[InvoiceList] = None,
-        commitment: Optional[Commitment] = Commitment.SINGLE
+        self, tx: solana.Transaction, invoice_list: Optional[InvoiceList] = None,
+        commitment: Optional[Commitment] = Commitment.SINGLE, dedupe_id: Optional[bytes] = None
     ) -> SubmitTransactionResult:
         """Submit a Solana transaction to Agora.
 
-        :param tx_bytes: The transaction, in bytes.
+        :param tx: The Solana transaction.
         :param invoice_list: (optional) An :class:`InvoiceList <agora.model.invoice.InvoiceList>` to associate with the
             transaction
+        :param commitment: The :class:`Commitment <agora.solana.commitment.Commitment>` to use.
+        :param dedupe_id: The dedupe ID to use for the transaction submission
         :return: A :class:`SubmitTransactionResult <agora.client.internal.SubmitTransactionResult>` object.
         """
 
         attempt = 0
+        tx_bytes = tx.marshal()
 
         def _submit():
             nonlocal attempt
@@ -359,6 +366,7 @@ class InternalClient:
                 ),
                 invoice_list=invoice_list.to_proto() if invoice_list else None,
                 commitment=commitment.to_proto(),
+                dedupe_id=dedupe_id,
             )
             resp = self._transaction_stub_v4.SubmitTransaction(req, metadata=self._metadata,
                                                                timeout=_GRPC_TIMEOUT_SECONDS)
@@ -376,7 +384,7 @@ class InternalClient:
                 if attempt == 1:
                     raise AlreadySubmittedError()
             elif resp.result == tx_pb_v4.SubmitTransactionResponse.Result.FAILED:
-                result.tx_error = TransactionErrors.from_proto_error(resp.transaction_error)
+                result.errors = TransactionErrors.from_solana_tx(tx, resp.transaction_error)
             elif resp.result == tx_pb_v4.SubmitTransactionResponse.Result.INVOICE_ERROR:
                 result.invoice_errors = resp.invoice_errors
             elif resp.result != tx_pb_v4.SubmitTransactionResponse.Result.OK:
